#!/bin/bash

# This script gets the version number of GMP.
# Assumes that CXX is set, and that a link to GMP header file is in ExternalLibs/include
# if there is no system default GMP.

# If an error occurs, exit code is non-zero (& mesg is printed on stderr).
# Otherwise exit code is zero, and output is version number (e.g. 6.0.1)

if [ "$#" != 0 ]
then
  echo "$0: ERROR: expects no args"
  exit 1
fi

if [ -z "$CXX" ]
then
  echo "$0: ERROR: shell variable CXX must be set to a C++ compiler compatible with GMP"
  exit 1
fi

if [ ! -d configuration ]
then
  echo "$0: ERROR: please run in the CoCoALib root directory"
  exit 1
fi

umask 22

# Get version number from the header file; we (ab)use the compiler.
# Create tmp directory, put C prog in it, compile, run, get output, delete directory.
# TMP_DIR depends on hostname, userid, and process number to try to avoid unfortunate
# name clashes if several people try to install CoCoALib simultaneously.
TMP_DIR=configuration/gmp-version-$UID@$HOSTNAME-$$
/bin/rm -rf $TMP_DIR && mkdir $TMP_DIR
if [ $? -ne 0 ]; then
  echobox "ERROR: $0: failed to create temporary directory \"$TMP_DIR\"";
  exit 1
fi
pushd $TMP_DIR >/dev/null
cat > TestProg.C <<EOF
#include "gmp.h"
#include <iostream>

int main()
{
  std::cout << __GNU_MP_VERSION << "." << __GNU_MP_VERSION_MINOR << "." << __GNU_MP_VERSION_PATCHLEVEL << std::endl;
}
EOF

# Use c++ compiler specified in CXX; no need to specify libgmp as all info is in header file!!
$CXX -I ../ExternalLibs/include  TestProg.C -o TestProg 2> /dev/null

# Check whether compilation failed; if so, complain.
if [ $? -ne 0 ]
then
  echo "ERROR: unable to determine version of GMP library"   > /dev/stderr
  echo "ERROR: (compilation failed in gmp-version.sh)"       > /dev/stderr
  popd
  /bin/rm -rf $TMP_DIR
  exit 1
fi

# Compilation succeeded, so run $PROG which will print out the version.
GMP_LIB_VERSION=`./TestProg`

popd >/dev/null
/bin/rm -rf $TMP_DIR

# If we get here, all tests have passed, so print version number and exit with code 0.
echo $GMP_LIB_VERSION
