//   Copyright (c)  2016 John Abbott, Anna M. Bigatti, Elisa Palezzato

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/MatrixOperations.H"
#include "CoCoA/CanonicalHom.H"
#include "CoCoA/DenseMatrix.H"
#include "CoCoA/FractionField.H"
#include "CoCoA/IntOperations.H"
#include "CoCoA/MatrixView.H"
#include "CoCoA/PolyRing.H"
#include "CoCoA/RingHom.H"
#include "CoCoA/RingZZ.H"
#include "CoCoA/apply.H"
#include "CoCoA/assert.H"
#include "CoCoA/bool3.H"
#include "CoCoA/convert.H"
#include "CoCoA/error.H"
#include "CoCoA/geobucket.H"
#include "CoCoA/time.H"
#include "CoCoA/utils.H" // for len

#include <algorithm>
using std::min;
#include <iostream>
using std::ostream;
#include <limits>
using std::numeric_limits;
//#include <vector>
using std::vector;

namespace CoCoA
{

  namespace // anonymous
  {

  class IncrLinDep
  {
  public:
    IncrLinDep(const ring& k, long dim);
    //    const std::vector<RingElem>& myAppendVec(std::vector<RingElem> v);
    void myAppendVec(std::vector<RingElem> v);
    const std::vector<RingElem>& myLinReln() {return myLinRelnValue;}
  
  private:
    ring myRing;
    int myVecLen;
    std::vector<std::vector<RingElem> > myM;
    std::vector<std::vector<RingElem> > myRowRepr;
    std::vector<int> myColTbl;
    std::vector<RingElem> myLinRelnValue;
  };

    IncrLinDep::IncrLinDep(const ring& K, long dim):
      myRing(K),
      myVecLen(dim)
    {
      if (!IsField(K)) CoCoA_ERROR(ERR::NotField, "IncrLinDep ctor");
      if (dim <= 0) CoCoA_ERROR(ERR::NotPositive, "IncrLinDep ctor");
      myM.reserve(dim);
      myRowRepr.reserve(dim);
      myColTbl.reserve(dim);
      myLinRelnValue.reserve(dim);
    }

  
  //  const std::vector<RingElem>& IncrLinDep::myAppendVec(std::vector<RingElem> v)
  void IncrLinDep::myAppendVec(std::vector<RingElem> v)
    {
      const long n = len(v);
      if (n != myVecLen) CoCoA_ERROR(ERR::IncompatDims, "IncrLinDep::myAppendVec");
      for (long i = 0; i < n; ++i)
        if (owner(v[i]) != myRing) CoCoA_ERROR(ERR::MixedRings, "IncrLinDep::myAppendVec");
      const long ncols = len(myM);
      myLinRelnValue.resize(ncols, zero(myRing));
      myLinRelnValue.push_back(one(myRing));
      for (long i=0; i < ncols; ++i)
      {
        const long col = myColTbl[i];
        if (IsZero(v[col])) continue;
        const RingElem c = v[col]/myM[i][col];
        for (long j=0; j < len(myRowRepr[i]); ++j)
          myLinRelnValue[j] -= c*myRowRepr[i][j];
        for (long j=0; j < n; ++j)
          v[j] -= c*myM[i][j];
      }
//      bool FoundLinReln = true;
      for (long i=0; i < n; ++i)
      {
        if (IsZero(v[i])) continue;
//        FoundLinReln = false;
        myM.push_back(v);
        myRowRepr.push_back(myLinRelnValue);
        myColTbl.push_back(i);
        myLinRelnValue.clear();
        break;
      }
      //      return myLinRelnValue;
    }


    RingElem CoeffOfTerm(ConstRefRingElem f, ConstRefPPMonoidElem pp)
    {
      for (SparsePolyIter itf=BeginIter(f); !IsEnded(itf); ++itf)
        if (PP(itf) == pp) return coeff(itf);
      return zero(CoeffRing(owner(f)));
    }


    void coefficients(vector<RingElem>& coeffs, ConstRefRingElem p, const vector<PPMonoidElem>& QB)
    {
      if (len(coeffs)!=len(QB))
        CoCoA_ERROR("len(coeffs)!=len(QB)","coefficients");
      for (long i=0; i<len(QB); ++i)  coeffs[i] = CoeffOfTerm(p, QB[i]);
    }


    void MatrixFlatten(vector<RingElem>& v, ConstMatrixView M)
    {
      if (len(v) != NumRows(M)*NumCols(M))
        CoCoA_ERROR("len(v) != NumRows(M)*NumCols(M)","MatrixFlatten");
      long k = -1;
      for (long i=0; i<NumRows(M); ++i)
        for (long j=0; j<NumCols(M); ++j)
          v[++k] = M(i,j);
    }
    
    
    matrix MultiplicationMat(ConstRefRingElem X, const ideal& I)
    {
      ring R = RingOf(I);
      if (R!=owner(X))
        CoCoA_ERROR("arguments must be in the same ring", "MultiplicationMat");
      if (!IsZeroDim(I))
        CoCoA_ERROR("ideal must be 0-dimensional", "MultiplicationMat");
      vector<PPMonoidElem> QB = QuotientBasis(I);
      //    std::cout << "len(QB) = " << len(QB) << std::endl;
      matrix M = NewDenseMat(CoeffRing(R), len(QB), len(QB));
      for (long j=0; j<len(QB); ++j)
      {
        RingElem p = NF(X*monomial(R, QB[j]), I);
        for (long i=0; i<len(QB); ++i)
          SetEntry(M, i, j, CoeffOfTerm(p, QB[i]));
      }
      return M;
    }

  } // anonymous namespace


  RingElem MinPoly(ConstMatrixView M, ConstRefRingElem x)
  {
    if (NumRows(M)!=NumCols(M))
      CoCoA_ERROR(ERR::NotSquareMatrix, "MinPoly");
    if (!IsIndet(x))
      CoCoA_ERROR("2nd arg must be an indeterminate","MinPoly");
  
    ring RM = RingOf(M);
    ring Rx = owner(x);
    if (RM==Rx)
      // TODO: convert M into matrix of coefficients (if possible)
      CoCoA_ERROR(ERR::NYI, "MinPoly");
    ring K = CoeffRing(Rx);
    //   If (IsZZ(R) or IsQQ(R)) Then
    //     M := apply(CanonicalHom(R, K), M);
    //     R := K;
    //   EndIf;
    if (K!=RM)
      CoCoA_ERROR(ERR::MixedRings, "MinPoly");
    long NRows = NumRows(M);
 
    IncrLinDep ILD(K, NRows*NRows);    
    vector<RingElem> flattenv(NRows*NRows, zero(K));

    matrix MD = NewDenseMat(IdentityMat(K, NRows));
    MatrixFlatten(flattenv, MD);
    ILD.myAppendVec(flattenv);

    double S1, S2, S5, V;
    while ( (ILD.myLinReln()).empty() )
    {
      V = CpuTime();
      MD = M*MD;
      S1 += (CpuTime()-V);
      V = CpuTime();
      MatrixFlatten(flattenv, MD);
      S2 += (CpuTime()-V);
      V = CpuTime();
      ILD.myAppendVec(flattenv);
      S5 += (CpuTime()-V);
    }

    double Y = CpuTime();
    RingHom phi = CoeffEmbeddingHom(Rx);
    RingElem mi = zero(owner(x));
    const vector<RingElem>& linreln = ILD.myLinReln();
    for (long d=0; d<len(linreln); ++d)  mi += phi(linreln[d])*power(x,d);
    Y = CpuTime() - Y;

    // std::cout << "Post-ciclo: " << Y << std::endl;
    // std::cout << "Somma M*MD: " << S1 << std::endl;
    // std::cout << "Somma flatten: " << S2 << std::endl;
    // std::cout << "Somma IncrLinDep: " << S5 << std::endl;
    // std::cout << "Somme totali: " << S1+S2+S5 << std::endl;
    return mi;
  }


  RingElem MinPolyMat2(ConstRefRingElem L, const ideal& I, ConstRefRingElem x)
  {
    return MinPoly(MultiplicationMat(L,I), x);
  }


  RingElem MinPolyMat(ConstRefRingElem L, const ideal& I, ConstRefRingElem x)
  {
    double U;
    U = CpuTime();
    matrix M = MultiplicationMat(L,I);
    // std::cout << "Tempo MultiplicationMat: " << (CpuTime() - U) << std::endl;

    U = CpuTime();
  
    //  If not(IsIndet(X)) Then error("MinPoly: 2nd arg must be an indeterminate"); EndIf;
  
    ring R = RingOf(M);
  
    //if R = RingOf(X) then return MinPolyInKx(M, X); endif;
  
    ring K = CoeffRing(owner(x));
    vector<PPMonoidElem> QB = QuotientBasis(I);
    IncrLinDep ILD(K, len(QB));
    
    vector<RingElem> flattenv(len(QB), zero(K));

    //  If R <> K Then error("MinPoly(M,x): rings of M and x are not compatible"); EndIf;

    //matrix MD = NewDenseMat(IdentityMat(K, NRows));
    matrix MD = NewDenseMat(ZeroMat(K, len(QB), 1));
    SetEntry(MD, 0,0, one(K));
    
    MatrixFlatten(flattenv, MD);
    ILD.myAppendVec(flattenv);
    // std::cout << "Pre-ciclo: " << (CpuTime() - U) << std::endl;

    U = CpuTime();
    double S1 = 0;
    double S2 = 0;
    double S5 = 0;  
    double V;
    while ( (ILD.myLinReln()).empty() )
    {
      V = CpuTime(); 
      MD = M*MD;
      S1 += (CpuTime()-V);
      V = CpuTime();
      MatrixFlatten(flattenv, MD);
      S2 += (CpuTime()-V);
      V = CpuTime();
      ILD.myAppendVec(flattenv);
      S5 += (CpuTime()-V);
    }
    // std::cout << "ciclo:" << (CpuTime() - U) << std::endl;

    double Y = CpuTime();
    RingHom phi = CoeffEmbeddingHom(owner(x));
    //ScalarProduct(col, [x^d | d in 0..i-1]);
    RingElem f = zero(owner(x));
    const vector<RingElem>& linreln = ILD.myLinReln();
    for (long d=0; d<len(linreln); ++d)  f += phi(linreln[d])*power(x,d);
    Y = CpuTime() - Y;

    // std::cout << "Post-ciclo: " << Y << std::endl;
    // std::cout << "Somma M*MD: " << S1 << std::endl;
    // std::cout << "Somma flatten: " << S2 << std::endl;
    // std::cout << "Somma IncrLinDep: " << S5 << std::endl;
    // std::cout << "Somme totali: " << S1+S2+S5 << std::endl;
    return f;
  }


RingElem MinPolyElim(ConstRefRingElem L, const ideal& I, ConstRefRingElem x)
{
  // controllare anelli
  ring R = RingOf(I);
  long n = NumIndets(R);
  //  myIndets = concat(IndetSymbols(R), [Record [head ="aux", indices =[]]]);
  ring S = NewPolyRing(CoeffRing(R), n+1);
  vector<RingElem> indetS_1;
  for (long i=0; i<n; ++i) indetS_1.push_back(indet(S,i));
  RingHom phi = PolyAlgebraHom(R, S, indetS_1);
  vector<RingElem> zero_x;
  for (long i=0; i<n; ++i) zero_x.push_back(zero(R));
  zero_x.push_back(x);
  RingHom psi = PolyAlgebraHom(S, R, zero_x);
  RingElem s = indet(S, n); // auxiliar variable
  ideal J = ideal(apply(phi, gens(I))) + ideal(s-phi(L));
  //  const double T = CpuTime();
  //  ideal El = elim(indetS_1, J);
  MakeUnique(J)->myElim(indetS_1);
  // std::cout << "Elim: " << CpuTime() -T << std::endl;
  return psi(monic(gens(J)[0]));
}


RingElem MinPolyDef(ConstRefRingElem r_orig, const ideal& I, ConstRefRingElem x)
{
  //  const double T = CpuTime();
  ring R = owner(r_orig);
  ring P = owner(x);
  ring K = CoeffRing(R);
  // fare tutti i controlli sugli anelli
  if (K != CoeffRing(P))
    CoCoA_ERROR("incompatible coeff rings", "MinPolyDef");
  RingElem r = NF(r_orig, I);
  vector<PPMonoidElem> QB = QuotientBasis(I);
  IncrLinDep ILD(K, len(QB));

    //  matrix M = NewDenseMat(K, len(QB), 1);
    //  SetEntry(M, 0,0, 1);

  vector<RingElem> coeffs(len(QB), zero(K));
  
  RingElem p = one(R);
  coeffs[0] = one(K);
  ILD.myAppendVec(coeffs);
  // std::cout << "Pre-ciclo: " << (CpuTime() - T) << std::endl;

  double W = CpuTime();
  double S1 = 0;
  double S2 = 0;
  double S3 = 0;
  double S4 = 0;
  double U;
  while ( (ILD.myLinReln()).empty() )
  {
    U = CpuTime();
    p = p*r;
    S1 = S1+(CpuTime()-U);

    U = CpuTime();
    p = NF(p, I);
    S2 = S2+(CpuTime()-U);
    
    U = CpuTime();
    coefficients(coeffs, p, QB);  // coefficients(p, QB);
    S3 = S3+(CpuTime()-U);

    U = CpuTime();
    ILD.myAppendVec(coeffs);
    S4 = S4+(CpuTime()-U);
  }
  W = CpuTime() - W;
  // std::cout << "ciclo: " << W << std::endl;

  double Y = CpuTime();
  RingHom phi = CoeffEmbeddingHom(P);
  //ScalarProduct(col, [x^d | d in 0..i-1]);
  RingElem f = zero(P);
  const vector<RingElem>& linreln = ILD.myLinReln();
  for (long d=0; d<len(linreln); ++d)  f += phi(linreln[d])*power(x,d);  
  Y = CpuTime() - Y;

  // std::cout << "Post-ciclo: " << Y << std::endl;
  // std::cout << "Somma interna al ciclo di: p^(i-1)*p --> " << S1 << std::endl;
  // std::cout << "Somma interna al ciclo di: NF --> " << S2 << std::endl;
  // std::cout << "Somma interna al ciclo di: coefficients --> " << S3 << std::endl;
  // std::cout << "Somma interna al ciclo di: IncrLinDep --> " << S4 << std::endl;
  // std::cout << "Somme totali interne al ciclo: " << S1+S2+S3+S4 << std::endl;
  return monic(f);
}

  

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/MinPoly.C,v 1.2 2016/03/18 15:13:47 abbott Exp $
// $Log: MinPoly.C,v $
// Revision 1.2  2016/03/18 15:13:47  abbott
// Summary: Commented out two unused variables (for timing)
//
// Revision 1.1  2016/03/18 11:54:00  bigatti
// -- first import
//
